/*
 * Exec denier (c) 2002 Kefren
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met: 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer. 2.
 * Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution. 3. All advertising
 * materials mentioning features or use of this software must display the
 * following acknowledgement: This product includes software developed by 
 * Mihai Chelaru. 4. My name may not be used to endorse
 * or promote products derived from this software without specific prior
 * written permission.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT I WILL BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/param.h>
#include <sys/ioctl.h>
#include <sys/systm.h>
#include <sys/conf.h>
#include <sys/mount.h>
#include <sys/exec.h>
#include <sys/lkm.h>
#include <sys/file.h>
#include <sys/errno.h>
#include <sys/syscall.h>
#include <sys/syscallargs.h>
#include <sys/sysctl.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <sys/stat.h>
#include <sys/fcntl.h>
#include <sys/timex.h>
#include <sys/mman.h>

#ifndef CURRENT
#include <sys/vnode.h>
#include <sys/socket.h>
#include <sys/socketvar.h>
#endif

#include "cfg.h"
#include "keflist.h"
#include "kefstr.h"
#include "myrealpath.h"
#include "kefinodelist.h"

static int      controlsyscall, defaultstate, mmapok;
struct elem    *l;
struct ielem   *il;
static struct sysent oldexec, oldmmap;

/* new exec syscall */
int
execcall(p, uap, retval)
	struct proc    *p;
	int            *uap;
	int             retval[];
{
	char           *comanda, rp[MAXPATHLEN];
	struct elem    *sl;
	int             denied = defaultstate, retbuf;
	comanda = (char *) uap[0];
	sl = seekuid(l, p->p_cred->p_ruid);
	myrealpath(comanda, rp, p);
	while (sl != NULL) {
		if (my_cmp(rp, sl->path, my_strlen(sl->path))) {
			denied = 0;
			break;
		}
		denied = 1;
		sl = seeknextuid(sl, p->p_cred->p_ruid);
	}
	if (denied) {
		printf("Denied: %s for UID: %d\n", comanda, p->p_cred->p_ruid);
		return EPERM;
	}
#ifdef VERBOSE
	printf("Allowed %s for UID: %d\n", comanda, p->p_cred->p_ruid);
#endif
	mmapok = TRUE;
	retbuf = oldexec.sy_call(p, uap, retval);
	mmapok = FALSE;
	return retbuf;
}

/* new mmap syscall */
int
mmapcall(p, uap, retval)
	struct proc    *p;
	struct sys_mmap_args *uap;
	int             retval[];
{
#ifdef MMAP_PARANOIA
	struct stat     ss;
	struct ielem   *isl;
	int             staterror;
	struct file    *fp;

	if ((SCARG(uap, prot) & PROT_EXEC) && (mmapok == FALSE)) {
		/* stat wrapper */
#ifdef CURRENT
		if ((fp = fd_getfile(p->p_fd, SCARG(uap, fd))) == NULL)
			return EBADF;
		FILE_USE(fp);
		staterror = (*fp->f_ops->fo_stat) (fp, &ss, p);
		FILE_UNUSE(fp, p);
		if (staterror != 0)
			return EPERM;
#else
		if (((uint) SCARG(uap, fd) >= p->p_fd->fd_nfiles) || ((fp = p->p_fd->fd_ofiles[SCARG(uap, fd)]) == NULL) || ((fp->f_iflags & FIF_WANTCLOSE) != 0))
			return EBADF;
		FILE_USE(fp);
		switch (fp->f_type) {
		case DTYPE_VNODE:
			staterror = vn_stat((struct vnode *) fp->f_data, &ss, p);
			break;
		case DTYPE_SOCKET:
			staterror = soo_stat((struct socket *) fp->f_data, &ss);
			break;
		default:
			panic("fstat wrapper error");
		}
		FILE_UNUSE(fp, p);
		if (staterror != 0)
			return EPERM;
#endif
		/* verifying */
		if (seekuid(l, p->p_cred->p_ruid) != NULL) {
			isl = iseekuid(il, p->p_cred->p_ruid);
			while (isl != NULL) {
				if ((isl->inode == ss.st_ino)&&(isl->device == ss.st_dev)) {
#ifdef VERBOSE
					printf("Allowed a mmap for UID: %d on inode: %u (fd: %d, device %d)\n", p->p_cred->p_ruid, ss.st_ino, SCARG(uap, fd), ss.st_dev);
#endif
					return oldmmap.sy_call(p, uap, retval);
				}
				isl = iseeknextuid(isl, p->p_cred->p_ruid);
			}
			printf("Denied a mmap for UID: %d on inode: %u (fd: %d, device: %d)\n", p->p_cred->p_ruid, ss.st_ino, SCARG(uap, fd), ss.st_dev);
			return EPERM;
		}
	}
#endif
	return oldmmap.sy_call(p, uap, retval);
}


/* Control syscall */
int
controlcall(p, uap, retval)
	struct proc    *p;
	int            *uap;
	int             retval[];
{
	int             f, uid;
	char           *str;
	struct elem    *sl;
	struct ielem   *isl;
	if (p->p_cred->p_ruid != 0)
		return EPERM;
	f = uap[0];
	uid = uap[1];
	str = (char *) uap[2];
	switch (f) {
	case 1:
		l = addelem(l, uid, str);
		if ((str[0] == '/') && (str[1] == 0))
			addelem(l, uid, ".");
		break;
	case 2:
		sl = seekuid(l, uid);
		while (sl != NULL) {
			l = delelem(sl);
			sl = seekuid(l, uid);
		}
		break;
	case 3:
		printall(l);
		break;
	case 4:
		defaultstate = uid;
		break;
	case 5:
		il = iaddelem(il, uid, uap[2], uap[3]);
		break;
	case 6:
		isl = iseekuid(il, uid);
		while (isl != NULL) {
			il = idelelem(isl);
			isl = iseekuid(il, uid);
		}
		break;
	case 7:
		iprintall(il);
		break;
	default:
		break;
	};
	return 0;
}

/* entries */

#ifdef CURRENT
static struct sysent execent = {
	3, 3 * sizeof(int), 1, execcall
};

static struct sysent controlent = {
	4, 4 * sizeof(int), 1, controlcall
};

static struct sysent mmapent = {
	7, 8 * sizeof(int), 1, mmapcall
};

#else
static struct sysent execent = {
	3, 3 * sizeof(int), execcall
};

static struct sysent controlent = {
	4, 4 * sizeof(int), controlcall
};

static struct sysent mmapent = {
	7, 8 * sizeof(int), mmapcall
};

#endif

/* Numele cacatului */
MOD_MISC("Denier")
/* Functia de incarcare si descarcare */

	static int
	                miscmod_handle(lkmtp, cmd)
	struct lkm_table *lkmtp;
	int             cmd;
{
	int             i;
	struct lkm_misc *args = lkmtp->private.lkm_misc;
	int             err = 0;/* default = success */
	extern int sys_lkmnosys __P((struct proc *, void *, register_t *));

	switch (cmd) {
	case LKM_E_LOAD:
		l = NULL;
		il = NULL;
		defaultstate = EXEC_PERMIT;
		mmapok = FALSE;
		if (lkmexists(lkmtp))
			return (EEXIST);

		/* Caut sloturi */
		for (i = 0; i < SYS_MAXSYSCALL - 4; i++)
			if (sysent[i].sy_call == sys_lkmnosys)
				break;	/* Evrika ! */
		/* Nashpa */
		if (i == SYS_MAXSYSCALL - 1) {
			err = ENFILE;
			break;
		}
		/* change exec */
		bcopy(&sysent[SYS_execve], &oldexec, sizeof(struct sysent));
		bcopy(&execent, &sysent[SYS_execve], sizeof(struct sysent));
		/* change mmap */
		bcopy(&sysent[SYS_mmap], &oldmmap, sizeof(struct sysent));
		bcopy(&mmapent, &sysent[SYS_mmap], sizeof(struct sysent));
		/* change control */
		bcopy(&controlent, &sysent[i], sizeof(struct sysent));
		controlsyscall = i;
		/* uraaaa */
		args->lkm_offset = i;	/* slot in sysent[] */
		printf("Control syscall is %d\n", controlsyscall);
		/* outta here */
		break;

	case LKM_E_UNLOAD:

		/* refac exec */
		bcopy(&oldexec, &sysent[SYS_execve], sizeof(struct sysent));
		/* refac mmap */
		bcopy(&oldmmap, &sysent[SYS_mmap], sizeof(struct sysent));
		/* eliberez syscallurile */
		sysent[controlsyscall].sy_call = sys_lkmnosys;
		/* list free */
		delall(l);
		idelall(il);
		/* die ! die ! die ! */
		break;

	default:		/* huh ? */
		err = EINVAL;
		break;
	}

	return err;
}

/* lkmentryu' */
int
denier_lkmentry(lkmtp, cmd, ver)
	struct lkm_table *lkmtp;
	int             cmd;
	int             ver;
{
	DISPATCH(lkmtp, cmd, ver, miscmod_handle, miscmod_handle, lkm_nofunc)
}

/* voila !!! */
